@@ -3,10 +3,16 @@
 # This source code is licensed under the MIT license found in the
 # LICENSE file in the root directory of this source tree.
 
-from .cost_function import AutoDiffCostFunction, CostFunction, ErrFnType
+from .cost_function import (
+    AutoDiffCostFunction,
+    AutogradMode,
+    CostFunction,
+    ErrFnType,
+    masked_jacobians,
+)
 from .cost_weight import CostWeight, DiagonalCostWeight, ScaleCostWeight
 from .objective import Objective
 from .robust_cost_function import RobustCostFunction
 from .robust_loss import HuberLoss, RobustLoss, WelschLoss
-from .variable import Variable
+from .variable import Variable, as_variable, masked_variables
 from .vectorizer import Vectorize
