@@ -4,6 +4,11 @@
 # LICENSE file in the root directory of this source tree.
 
 from .lie_group import LieGroup, adjoint, between, compose, exp_map, inverse, log_map
+from .lie_group_check import (
+    enable_lie_group_check,
+    no_lie_group_check,
+    set_lie_group_check_enabled,
+)
 from .manifold import Manifold, OptionalJacobians, local, retract
 from .point_types import (
     Point2,
